/* Dalloc - "Debugging" malloc library for ANSI C.
          - Written by Jeff Hay, jrhay@lanl.gov
*/
#include "dalloc.h"
#include "rsync.h"
#undef malloc
#undef free
#undef strdup
#undef realloc
#undef calloc

#ifdef __NETWARE__
//DmemPtr DmemRoot = NULL;

DmemPtr NewDmem(void *ID, size_t size)
{
  get_tsd
  DmemPtr NewPtr, Prev, Track;

  if ((NewPtr = DNEXT_MALLOC(sizeof(struct Dmem))) != NULL)
    {
      NewPtr->ID = ID;
      NewPtr->size = size;
      NewPtr->Next = NULL;

	if (DmemRoot == NULL)
	{
		DmemRoot = NewPtr;
//		if (verbose >= 1)
//			consoleprintf("%d NewDmem DmemRoot=%x\n", getpid(), DmemRoot);
	}
	else
	{
	  Prev = Track = DmemRoot;
	  while ((Track != NULL) && (Track->Next != NULL))
	    {
	      Prev = Track;
	      Track = Track->Next;
	    }
	  if (Track != NULL)
	    Track->Next = NewPtr;
	  else
	    Prev->Next = NewPtr;
	}
    }
  
  return NewPtr;
} /* NewDmem() */    

DmemPtr FindDmem(void *ID, DmemPtr *APrev)
{
  get_tsd
  DmemPtr Track, Prev;

  Prev = NULL;
  Track = DmemRoot;
  while ((Track != NULL) && (Track->ID != ID))
    {
      Prev = Track;
      Track = Track->Next;
    }

  if (APrev != NULL)
	  *APrev = Prev;

  return Track;
} /* FindDmem() */

void *dalloc(size_t size, const char* file, int line)
{
  get_tsd
  void *NewMem;
  DmemPtr NewPtr;

  if ((NewMem = DNEXT_MALLOC(size)) == NULL)
    return NULL;

  if ((NewPtr = NewDmem(NewMem, size)) == NULL)
    {
      DNEXT_FREE(NewMem);
      return NULL;
    }
    
	//if (verbose >= 1)
	//	consoleprintf("%x==Alloc(%d) (%s,%d)\n", NewMem, size, file, line);

  return NewMem;
} /* dalloc() */

void *dcalloc(size_t nmemb, size_t size, const char* file, int line)
{
  get_tsd
  void *NewMem;
  DmemPtr NewPtr;

  if ((NewMem = DNEXT_CALLOC(nmemb, size)) == NULL)
    return NULL;

  if ((NewPtr = NewDmem(NewMem, nmemb*size)) == NULL)
    {
      DNEXT_FREE(NewMem);
      return NULL;
    }

	//if (verbose >= 1)
	//	consoleprintf("%x==Alloc(%d) (%s,%d)\n", NewMem, NewPtr->size, file, line);
  return NewMem;
} /* dcalloc() */

void *drealloc(void *ID, size_t size, const char* file, int line)
{
  get_tsd	
  DmemPtr Old;
  void *New;
  void *oldID = NULL;
  size_t oldSize = 0;

  if (size == 0) 
    {
      dfree(ID,file,line);
      return NULL;
    }
  if (ID == NULL)
    return dalloc(size,file,line);

  Old = FindDmem(ID, NULL);
  if ((New = DNEXT_REALLOC(ID, size)) != NULL)
    if (New != NULL) 
      {
     oldID = Old->ID;
     oldSize = Old->size; 
	 Old->ID = New;
	 Old->size = size;
      }

    if (ID != New)  
    {
		//if (verbose >= 1)
		//{
		//consoleprintf("%x==ReAlloc(%d,%x) (%s,%d)\n", New, size, ID,file, line);
		//consoleprintf("%x==Free(%d) (%s,%d)\n", oldID, size, file, line);
		//}
	}		
  return New;
} /* drealloc() */

char *dstrdup(const char * str, const char* file, int line)
{
	get_tsd
	void *NewMem;
	DmemPtr NewPtr;
	int size = strlen(str)+1;

	if ((NewMem = DNEXT_MALLOC(size)) == NULL)
	return NULL;

	if ((NewPtr = NewDmem(NewMem, size)) == NULL)
	{
	  DNEXT_FREE(NewMem);
	  return NULL;
	}

	strcpy(NewMem,str);
	//if (verbose >= 1)
	//	consoleprintf("%x==StrAlloc(%d) (%s,%d)\n", NewMem, size, file, line);
	return NewMem;
} /* dcalloc() */

void dfree(void *ID, const char* file, int line) 
{
	get_tsd
	DmemPtr Track, Prev;

	if ((Track = FindDmem(ID, &Prev)) != NULL)
	{
		if (Prev != NULL)
			Prev->Next = Track->Next;
		else
		{
			DmemRoot = Track->Next;
			//if (verbose >= 1)
			//	consoleprintf("%d dfree DmemRoot=%x\n", getpid(), DmemRoot);
		}
		//if (verbose >= 1)
		//	consoleprintf("%x==Free(%d) (%s,%d)\n", Track->ID, Track->size, file, line);
	
	     DNEXT_FREE(Track);
  }
  else
  {
  	//if (verbose >= 1)
	//	consoleprintf("%x==dfree() (%s,%d) did not find\n", ID, file, line);
  }
  DNEXT_FREE(ID);

  return;
} /* dfree() */

long dcount(long *Units) 
{
  get_tsd
  long Count = 0;
  long UnitCount = 0;
  DmemPtr Track;

  Track = DmemRoot;
  while (Track != NULL)
    {
      Count += Track->size;
      UnitCount++;
      Track = Track->Next;
    }

  if (Units != NULL)
    *Units = UnitCount;

  return Count;
} /* dcount() */

void dfreeall() 
{
    get_tsd
	DmemPtr Track;
	//if (verbose >= 1)
	//	consoleprintf("dfreeall %x\n", DmemRoot);

	Track = DmemRoot;
	while (Track != NULL)
	{
		DmemPtr Next = Track->Next;
		//if (verbose >= 1)
		//	consoleprintf("%x==Free(%d)\n", Track->ID, Track->size);
		DNEXT_FREE(Track->ID);
		DNEXT_FREE(Track);
	  Track = Next;
	}
	DmemRoot = NULL;

} /* dcount() */

#endif /* DEBUG */

#ifdef __TEST_DALLOC__
#include <stdio.h>
#include <stdlib.h>

#define DALLOC_USED used = DCOUNT(&allocs); \
           printf("%d bytes allocated in %d allocations\n", used, allocs);

int main()
{
  long used, allocs;
  int *memory1, *memory2;

  printf("dalloc - Debugging Memory Allocator for ANSI C.\n\n");

  printf("Allocating 1 int....\n");
  memory1 = DMALLOC(sizeof(int));
  DALLOC_USED
  
  printf("Allocating array of 20 ints...\n");
  memory2 = DCALLOC(20, sizeof(int));
  DALLOC_USED

  printf("Re-sizing single int to 10 ints...\n");
  memory1 = DREALLOC(memory1, 10*sizeof(int));
  DALLOC_USED

//  printf("Freeing array of 20 ints...\n");
//  DFREE(memory2);
//  DALLOC_USED

  printf("Re-sizing 10 int space to 0 (hence, freeing)...\n");
//  memory1 = DREALLOC(memory1, 0);
  DALLOC_USED

  printf("dalloc demo complete.\n");
//dfreeall();
  return 0;
} /* main() */
#endif /* __TEST_DALLOC__ */  




