/* 
   Copyright (C) by Andrew Tridgell 1996, 2000
   Copyright (C) Paul Mackerras 1996
   Copyright (C) 2001, 2002 by Martin Pool <mbp@samba.org>
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#define False 0
#define True 1

#define BLOCK_SIZE 700
#define RSYNC_RSH_ENV "RSYNC_RSH"
#define RSYNC_RSH_IO_ENV "RSYNC_RSH_IO"

#define RSYNC_NAME "rsync"
/* RSYNCD_SYSCONF is now set in config.h */
#define RSYNCD_USERCONF "rsyncd.conf"

#define DEFAULT_LOCK_FILE "/var/run/rsyncd.lock"
#define URL_PREFIX "rsync://"

#define BACKUP_SUFFIX "~"

/* a non-zero CHAR_OFFSET makes the rolling sum stronger, but is
   incompatible with older versions :-( */
#define CHAR_OFFSET 0

/* These flags are only used during the flist transfer. */

#define XMIT_TOP_DIR (1<<0)
#define XMIT_SAME_MODE (1<<1)
#define XMIT_EXTENDED_FLAGS (1<<2)
#define XMIT_SAME_RDEV_pre28 XMIT_EXTENDED_FLAGS /* Only in protocols < 28 */
#define XMIT_SAME_UID (1<<3)
#define XMIT_SAME_GID (1<<4)
#define XMIT_SAME_NAME (1<<5)
#define XMIT_LONG_NAME (1<<6)
#define XMIT_SAME_TIME (1<<7)
#define XMIT_SAME_RDEV_MAJOR (1<<8)
#define XMIT_HAS_IDEV_DATA (1<<9)
#define XMIT_SAME_DEV (1<<10)
#define XMIT_RDEV_MINOR_IS_SMALL (1<<11)

/* These flags are used in the live flist data. */

#define FLAG_TOP_DIR (1<<0)
#define FLAG_HLINK_EOL (1<<1)	/* generator only */
#define FLAG_MOUNT_POINT (1<<2)	/* sender only */

/* update this if you make incompatible changes */
#define PROTOCOL_VERSION 28

/* We refuse to interoperate with versions that are not in this range.
 * Note that we assume we'll work with later versions: the onus is on
 * people writing them to make sure that they don't send us anything
 * we won't understand.
 *
 * Interoperation with old but supported protocol versions
 * should cause a warning to be printed.  At a future date
 * the old protocol will become the minimum and
 * compatibility code removed.
 *
 * There are two possible explanations for the limit at
 * MAX_PROTOCOL_VERSION: either to allow new major-rev versions that
 * do not interoperate with us, and (more likely) so that we can
 * detect an attempt to connect rsync to a non-rsync server, which is
 * unlikely to begin by sending a byte between MIN_PROTOCL_VERSION and
 * MAX_PROTOCOL_VERSION. */

#define MIN_PROTOCOL_VERSION 20
#define OLD_PROTOCOL_VERSION 25
#define MAX_PROTOCOL_VERSION 40

#define RSYNC_PORT 873

#define SPARSE_WRITE_SIZE (1024)
#define WRITE_SIZE (32*1024)
#define CHUNK_SIZE (32*1024)
#define MAX_MAP_SIZE (256*1024)
#define IO_BUFFER_SIZE (4092)

#define IOERR_GENERAL	(1<<0) /* For backward compatibility, this must == 1 */
#define IOERR_VANISHED	(1<<1)

#define MAX_ARGS 1000

#define MPLEX_BASE 7

#define NO_EXCLUDES	0
#define SERVER_EXCLUDES	1
#define ALL_EXCLUDES	2

#define XFLG_FATAL_ERRORS	(1<<0)
#define XFLG_DEF_INCLUDE	(1<<1)
#define XFLG_WORDS_ONLY 	(1<<2)
#define XFLG_WORD_SPLIT 	(1<<3)
#define XFLG_DIRECTORY	 	(1<<4)

#define PERMS_REPORT		(1<<0)
#define PERMS_SKIP_MTIME	(1<<1)

#define FULL_FLUSH	1
#define NORMAL_FLUSH	0

#define PDIR_CREATE	1
#define PDIR_DELETE	0


/* Log-message categories.  FLOG is only used on the daemon side to
 * output messages to the log file. */
enum logcode { FERROR=1, FINFO=2, FLOG=3 };

/* Messages types that are sent over the message channel.  The logcode
 * values must all be present here with identical numbers. */
enum msgcode {
	MSG_DONE=5,	/* current phase is done */
	MSG_REDO=4,	/* reprocess indicated flist index */
	MSG_ERROR=FERROR, MSG_INFO=FINFO, MSG_LOG=FLOG, /* remote logging */
	MSG_DATA=0	/* raw data on the multiplexed stream */
};

#include "errcode.h"

#include "config.h"

/* The default RSYNC_RSH is always set in config.h. */

#include <sys/types.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#include <stddef.h>

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if defined(HAVE_MALLOC_H) && (defined(HAVE_MALLINFO) || !defined(HAVE_STDLIB_H))
#include <malloc.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#endif

#ifdef __NETWARE__
#define _POSIX_SOURCE
#endif
#include <sys/stat.h>

#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

#include <signal.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_GRP_H
#include <grp.h>
#endif
#include <errno.h>

#ifdef HAVE_UTIME_H
#include <utime.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_SYS_MODE_H
/* apparently AIX needs this for S_ISLNK */
#ifndef S_ISLNK
#include <sys/mode.h>
#endif
#endif

#ifdef HAVE_GLOB_H
#include <glob.h>
#endif

/* these are needed for the uid/gid mapping code */
#include <pwd.h>
#include <grp.h>

#include <stdarg.h>
#ifndef __NETWARE__
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#ifdef __NETWARE__

#define WINSOCK
#include "zlib/zlib.h"

//#include <ws2nlm.h>
#include <novsock2.h>

#include <proc.h>
#include <netware.h>
#include <library.h>
#include <screen.h>
#include <nks/plat.h>
#endif

#include <syslog.h>
#include <sys/file.h>

#ifdef __NETWARE__
// Just a few defines to get things to compile on NetWare
typedef int socklen_t;

enum threadtype {TMAIN=0, TGEN=1, TRECV=2 };

//#define NW_WAIT(x) { char g[10]; printf("\n%s\nPress ENTER to continue:",(x)); gets(g); }
#define NW_WAIT(x) ;

#define SHUTDOWN(s) { shutdown(s,SD_BOTH);closesocket(s);}

#define C_SHUTDOWN(s) { \
 char bfr; \
 shutdown(s, 1); \
 while( recv( s, &bfr, 1, 0 ) > 0 ); \
 shutdown(s, 0); \
 closesocket( s );}

//#define DEBUG
#endif

#if HAVE_DIRENT_H
# include <dirent.h>
#else
# define dirent direct
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#if MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#elif MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif

#ifdef HAVE_COMPAT_H
#include <compat.h>
#endif

#include <assert.h>

#include "lib/pool_alloc.h"

#ifndef WINSOCK
#define BOOL int
#endif

#ifndef uchar
#define uchar unsigned char
#endif

#if HAVE_UNSIGNED_CHAR
#define schar signed char
#else
#define schar char
#endif

#ifndef int32
#if (SIZEOF_INT == 4)
#define int32 int
#elif (SIZEOF_LONG == 4)
#define int32 long
#elif (SIZEOF_SHORT == 4)
#define int32 short
#else
/* I hope this works */
#define int32 int
#define LARGE_INT32
#endif
#endif

#ifndef uint32
#define uint32 unsigned int32
#endif

#if HAVE_OFF64_T
#define OFF_T off64_t
#define STRUCT_STAT struct stat64
#else
#define OFF_T off_t
#define STRUCT_STAT struct stat
#endif

#if HAVE_OFF64_T
#define int64 off64_t
#elif (SIZEOF_LONG == 8) 
#define int64 long
#elif (SIZEOF_INT == 8) 
#define int64 int
#elif HAVE_LONGLONG
#if defined( __NETWARE__ ) && defined(WATCOM)
#define int64 __int64
#else
#define int64 long long
#endif
#else
/* As long as it gets... */
#define int64 off_t
#define INT64_IS_OFF_T
#endif

#if (SIZEOF_LONG == 8) 
#define uint64 unsigned long
#elif (SIZEOF_INT == 8) 
#define uint64 unsigned int
#elif HAVE_LONGLONG
#if defined( __NETWARE__ ) && defined(WATCOM)
#define uint64 unsigned __int64
#else
#define uint64 unsigned long long
#endif
#else
/* As long as it gets... */
#define uint64 unsigned off_t
#endif

/* Starting from protocol version 26, we always use 64-bit
 * ino_t and dev_t internally, even if this platform does not
 * allow files to have 64-bit inums.  That's because the
 * receiver needs to find duplicate (dev,ino) tuples to detect
 * hardlinks, and it might have files coming from a platform
 * that has 64-bit inums.
 *
 * The only exception is if we're on a platform with no 64-bit type at
 * all.
 *
 * Because we use read_longint() to get these off the wire, if you
 * transfer devices or hardlinks with dev or inum > 2**32 to a machine
 * with no 64-bit types then you will get an overflow error.  Probably
 * not many people have that combination of machines, and you can
 * avoid it by not preserving hardlinks or not transferring device
 * nodes.  It's not clear that any other behaviour is better.
 *
 * Note that if you transfer devices from a 64-bit-devt machine (say,
 * Solaris) to a 32-bit-devt machine (say, Linux-2.2/x86) then the
 * device numbers will be truncated.  But it's a kind of silly thing
 * to do anyhow.
 *
 * FIXME: I don't think the code in flist.c has ever worked on a system
 * where dev_t is a struct.
 */ 

struct idev {
	uint64 inode;
	uint64 dev;
};

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif

/* the length of the md4 checksum */
#define MD4_SUM_LENGTH 16
#define SUM_LENGTH 16
#define SHORT_SUM_LENGTH 2
#define BLOCKSUM_BIAS 10

#ifndef MAXPATHLEN
#ifdef __NETWARE__
#define MAXPATHLEN PATH_MAX
#else
#define MAXPATHLEN 1024
#endif
#endif

#ifndef NAME_MAX
#define NAME_MAX 255
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

#ifndef IN_LOOPBACKNET
#define IN_LOOPBACKNET 127
#endif

#define GID_NONE ((gid_t)-1)

#define HL_CHECK_MASTER	0
#define HL_SKIP		1

struct hlink {
	int hlindex;
	struct file_struct *next;
};

#define F_DEV	link_u.idev->dev
#define F_INODE	link_u.idev->inode

#define F_HLINDEX link_u.links->hlindex
#define F_NEXT	link_u.links->next

struct file_struct {
	union {
		dev_t rdev;	/* The device number, if this is a device */
		char *sum;	/* Only a normal file can have a checksum */
		char *link;	/* Points to symlink string, if a symlink */
	} u;
	OFF_T length;
	char *basename;
	char *dirname;
	char *basedir;
	union {
		struct idev *idev;
		struct hlink *links;
	} link_u;
	time_t modtime;
	uid_t uid;
	gid_t gid;
	mode_t mode;
	uchar flags;	/* this item MUST remain last */
};

/*
 * Start the flist array at FLIST_START entries and grow it
 * by doubling until FLIST_LINEAR then grow by FLIST_LINEAR
 */
#define FLIST_START	(32 * 1024)
#define FLIST_LINEAR	(FLIST_START * 512)

/*
 * Extent size for allocation pools A minimum size of 128KB
 * is needed to mmap them so that freeing will release the
 * space to the OS.
 *
 * Larger sizes reduce leftover fragments and speed free calls
 * (when they happen) Smaller sizes increase the chance of
 * freed allocations freeing whole extents.
 */
#define FILE_EXTENT	(256 * 1024)
#define HLINK_EXTENT	(128 * 1024)

#define WITH_HLINK	1
#define WITHOUT_HLINK	0

struct file_list {
	int count;
	int malloced;
	alloc_pool_t file_pool;
	alloc_pool_t hlink_pool;
	struct file_struct **files;
};

#define SUMFLG_SAME_OFFSET	(1<<0)

struct sum_buf {
	OFF_T offset;		/**< offset in file of this chunk */
	unsigned int len;	/**< length of chunk of file */
	uint32 sum1;	        /**< simple checksum */
	short flags;		/**< flag bits */
	char sum2[SUM_LENGTH];	/**< checksum  */
};

struct sum_struct {
	OFF_T flength;		/**< total file length */
	size_t count;		/**< how many chunks */
	unsigned int blength;	/**< block_length */
	unsigned int remainder;	/**< flength % block_length */
	int s2length;		/**< sum2_length */
	struct sum_buf *sums;	/**< points to info for each chunk */
};

struct map_struct {
	char *p;		/* Window pointer			*/
	int fd;			/* File Descriptor			*/
	int p_size;		/* Largest window size we allocated	*/
	int p_len;		/* Latest (rounded) window size		*/
	int def_window_size;	/* Default window size			*/
	int status;		/* first errno from read errors		*/
	OFF_T file_size;	/* File size (from stat)		*/
	OFF_T p_offset;		/* Window start				*/
	OFF_T p_fd_offset;	/* offset of cursor in fd ala lseek	*/
};

#define MATCHFLG_WILD		(1<<0) /* pattern has '*', '[', and/or '?' */
#define MATCHFLG_WILD2		(1<<1) /* pattern has '**' */
#define MATCHFLG_WILD2_PREFIX	(1<<2) /* pattern starts with '**' */
#define MATCHFLG_ABS_PATH	(1<<3) /* path-match on absolute path */
#define MATCHFLG_INCLUDE	(1<<4) /* this is an include, not an exclude */
#define MATCHFLG_DIRECTORY	(1<<5) /* this matches only directories */
#define MATCHFLG_CLEAR_LIST 	(1<<6) /* this item is the "!" token */
struct exclude_struct {
	struct exclude_struct *next;
	char *pattern;
	unsigned int match_flags;
	int slash_cnt;
};

struct exclude_list_struct {
	struct exclude_struct *head;
	struct exclude_struct *tail;
	char *debug_type;
};

struct stats {
	int64 total_size;
	int64 total_transferred_size;
	int64 total_written;
	int64 total_read;
	int64 literal_data;
	int64 matched_data;
	int flist_size;
	int num_files;
	int num_transferred_files;
	int current_file_index;
};


#ifdef __NETWARE__
struct jni_stats {
	int64 total_size;
	int64 total_transferred_size;
	int64 total_written;
	int64 total_read;
	int64 literal_data;
	int64 matched_data;
	int64 flist_size;
	int64 num_files;
	int64 num_transferred_files;
	time_t jni_starttime;
	int64 flist_index;
	char op[5];
	char fname[MAXPATHLEN];
};
#endif
/* we need this function because of the silly way in which duplicate
   entries are handled in the file lists - we can't change this
   without breaking existing versions */
#if defined( __NETWARE__ ) && defined(WATCOM)
static __inline int flist_up(struct file_list *flist, int i)
#else
static inline int flist_up(struct file_list *flist, int i)
#endif
{
	while (!flist->files[i]->basename) i++;
	return i;
}

#ifdef __NETWARE__
#include "bytorder.h"
#else
#include "byteorder.h"
#endif
#include "lib/mdfour.h"
#include "lib/wildmatch.h"
#include "lib/permstring.h"
#include "lib/addrinfo.h"

#include "proto.h"

/* We have replacement versions of these if they're missing. */
#ifndef HAVE_ASPRINTF
int asprintf(char **ptr, const char *format, ...);
#endif

#ifndef HAVE_VASPRINTF
int vasprintf(char **ptr, const char *format, va_list ap);
#endif

#if !defined(HAVE_VSNPRINTF) || !defined(HAVE_C99_VSNPRINTF)
#define vsnprintf rsync_vsnprintf
int vsnprintf(char *str, size_t count, const char *fmt, va_list args);
#endif

#if !defined(HAVE_SNPRINTF) || !defined(HAVE_C99_VSNPRINTF)
#define snprintf rsync_snprintf
int snprintf(char *str,size_t count,const char *fmt,...);
#endif


#if !HAVE_STRERROR
extern char *sys_errlist[];
#define strerror(i) sys_errlist[i]
#endif

#ifndef HAVE_STRCHR
# define strchr                 index
# define strrchr                rindex
#endif

#ifndef HAVE_ERRNO_DECL
extern int errno;
#endif

#define SUPPORT_LINKS HAVE_READLINK
#define SUPPORT_HARD_LINKS HAVE_LINK

/* This could be bad on systems which have no lchown and where chown
 * follows symbollic links.  On such systems it might be better not to
 * try to chown symlinks at all. */
#ifndef HAVE_LCHOWN
#define lchown chown
#endif

#define SIGNAL_CAST (RETSIGTYPE (*)())

#ifndef EWOULDBLOCK
#define EWOULDBLOCK EAGAIN
#endif

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif

#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif

#ifndef STDERR_FILENO
#define STDERR_FILENO 2
#endif

#ifndef S_IRUSR
#define S_IRUSR 0400
#endif

#ifndef S_IWUSR
#define S_IWUSR 0200
#endif

#ifndef ACCESSPERMS
#define ACCESSPERMS 0777
#endif

#ifndef S_ISVTX
#define S_ISVTX 0
#endif

#define CHMOD_BITS (S_ISUID | S_ISGID | S_ISVTX | ACCESSPERMS)

#ifndef _S_IFMT
#define _S_IFMT        0170000
#endif

#ifndef _S_IFLNK
#define _S_IFLNK  0120000
#endif

#ifndef S_ISLNK
#define S_ISLNK(mode) (((mode) & (_S_IFMT)) == (_S_IFLNK))
#endif

#ifndef S_ISBLK
#define S_ISBLK(mode) (((mode) & (_S_IFMT)) == (_S_IFBLK))
#endif

#ifndef S_ISCHR
#define S_ISCHR(mode) (((mode) & (_S_IFMT)) == (_S_IFCHR))
#endif

#ifndef S_ISSOCK
#ifdef _S_IFSOCK
#define S_ISSOCK(mode) (((mode) & (_S_IFMT)) == (_S_IFSOCK))
#else
#define S_ISSOCK(mode) (0)
#endif
#endif

#ifndef S_ISFIFO
#ifdef _S_IFIFO
#define S_ISFIFO(mode) (((mode) & (_S_IFMT)) == (_S_IFIFO))
#else
#define S_ISFIFO(mode) (0)
#endif
#endif

#ifndef S_ISDIR
#define S_ISDIR(mode) (((mode) & (_S_IFMT)) == (_S_IFDIR))
#endif

#ifndef S_ISREG
#define S_ISREG(mode) (((mode) & (_S_IFMT)) == (_S_IFREG))
#endif

/* work out what fcntl flag to use for non-blocking */
#ifdef O_NONBLOCK
# define NONBLOCK_FLAG O_NONBLOCK
#elif defined(SYSV)
# define NONBLOCK_FLAG O_NDELAY
#else 
# define NONBLOCK_FLAG FNDELAY
#endif

#ifndef INADDR_LOOPBACK
#define INADDR_LOOPBACK 0x7f000001
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

#define IS_DEVICE(mode) (S_ISCHR(mode) || S_ISBLK(mode) || S_ISSOCK(mode) || S_ISFIFO(mode))

/* Initial mask on permissions given to temporary files.  Mask off setuid
     bits and group access because of potential race-condition security
     holes, and mask other access because mode 707 is bizarre */
#define INITACCESSPERMS 0700

/* handler for null strings in printf format */
#define NS(s) ((s)?(s):"<NULL>")

#if !defined(__GNUC__) || defined(APPLE)
/* Apparently the OS X port of gcc gags on __attribute__.
 *
 * <http://www.opensource.apple.com/bugs/X/gcc/2512150.html> */
#define __attribute__(x) 

#endif

/* Convenient wrappers for malloc and realloc.  Use them. */
#define new(type) ((type *)malloc(sizeof(type)))
#define new_array(type, num) ((type *)_new_array(sizeof(type), (num)))
#define realloc_array(ptr, type, num) ((type *)_realloc_array((ptr), sizeof(type), (num)))

/* use magic gcc attributes to catch format errors */
 void rprintf(enum logcode , const char *, ...)
     __attribute__((format (printf, 2, 3)))
;

/* This is just like rprintf, but it also tries to print some
 * representation of the error code.  Normally errcode = errno. */
void rsyserr(enum logcode, int, const char *, ...)
     __attribute__((format (printf, 3, 4)))
     ;

#ifdef REPLACE_INET_NTOA
#define inet_ntoa rep_inet_ntoa
#endif

/* Make sure that the O_BINARY flag is defined. */
#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *d, const char *s, size_t bufsize);
#endif

#ifndef HAVE_STRLCAT
size_t strlcat(char *d, const char *s, size_t bufsize);
#endif

#ifndef WEXITSTATUS
#define	WEXITSTATUS(stat)	((int)(((stat)>>8)&0xFF))
#endif

#define exit_cleanup(code) _exit_cleanup(code, __FILE__, __LINE__)

#ifdef HAVE_GETEUID
#define MY_UID() geteuid()
#else
#define MY_UID() getuid()
#endif

#ifdef HAVE_GETEGID
#define MY_GID() getegid()
#else
#define MY_GID() getgid()
#endif

#ifdef __NETWARE__
#define SCOREBOARD
#ifdef SCOREBOARD
#include "scoreboard.h"
#endif

#define NW_CSUM_CHUNK 64

#define makedev(major, minor) (((major) << 8) | ((minor) & 0xff))
#define major(dev) ((int)(((dev) >> 16) & 0xffff))
#define minor(dev) ((int)((dev) & 0xffff))

#define PROGRESS_HISTORY_SECS 5

struct progress_history {
	struct timeval time;
	OFF_T ofs;
};

struct redo_list {
	struct redo_list *next;
	int num;
};

struct msg_list {
	struct msg_list *next;
	char *buf;
	int len;
};

typedef void mutex;
typedef void semaphore;
typedef void thread;
typedef void event;
typedef struct stats nwstats;

extern int volatile m_NLM_shutdown;
extern int volatile m_NLM_threadCount;
extern semaphore * m_RecvSemaphore;
extern semaphore * m_NamedSemaphore;
extern int m_mainThreadGroupID;
extern int gInited;
extern scr_t gScreenHandle;
extern int g_ssl;
/*
 * This structure describes a single service.
 */
 
#include <pthread.h>
 
typedef struct
{
	char *name;
	char *path;
	char *comment;
	char *lock_file;
	BOOL read_only;
	BOOL write_only;
	BOOL list;
	BOOL use_chroot;
	BOOL transfer_logging;
	BOOL ignore_errors;
	char *uid;
	char *gid;
	char *hosts_allow;
	char *hosts_deny;
	char *auth_users;
	char *secrets_file;
	BOOL strict_modes;
	char *exclude;
	char *exclude_from;
	char *include;
	char *include_from;
	char *log_format;
	char *refuse_options;
	char *dont_compress;
	int timeout;
	int max_connections;
	BOOL ignore_nonreadable;
	BOOL live_file_system;
} service;

/* Thread Storage Data */
typedef struct _TSD {
// These are the externs put into tls
	int accept_fd;
	void *threadParams;
	BOOL bReleased;
	int rsync_argc;
	char ** rsync_argv;

	// Warning, keep these in this order for the stream flags,
	// check batch.c	
	int recurse;
	int preserve_uid;
	int preserve_gid;	
	int preserve_links;
	int preserve_devices;
	int preserve_hard_links;	
	int always_checksum;	
	int *flag_ptr[8];
	
	int io_multiplexing_out;
	int io_multiplexing_in;
	int multiplex_in_fd;
	int multiplex_out_fd;
	int sock_f_in;
	int sock_f_out;
	int write_batch_monitor_in;
	int write_batch_monitor_out;
	time_t last_io;
	int no_flush;
	char *iobuf_out;
	int iobuf_out_cnt;
	char *iobuf_in;
	size_t iobuf_in_siz;	
	size_t iobuf_in_ndx;
	char full_name[MAXPATHLEN];
	char curr_dir[MAXPATHLEN];
	unsigned int curr_dir_len;
	char backup_dir_buf[MAXPATHLEN];
	unsigned int backup_dir_remainder;
	int log_got_error;
	int verbose;
	int quiet;
	int make_backups;
	int whole_file;
	int no_whole_file;
	int copy_links;
	int preserve_perms;
	int preserve_times;
	int update_only;
	int cvs_exclude;
	int dry_run;
	int local_server;
	int ignore_times;
	int delete_mode;
	int delete_excluded;
	int one_file_system;
	int protocol_version;
	int remote_protocol;
	int sparse_files;
	int do_compression;
	int compression_level;
	int am_root;
	int orig_umask;
	int relative_paths;
	int numeric_ids;
	int force_delete;
	int io_timeout;
	int io_error;
	int read_only;
	int module_id;
	int am_server;
	int am_sender;
	int am_generator;
	int daemon_over_rsh;
	int do_stats;
	int do_progress;
	int keep_partial;
	int select_timeout;
	char *partial_dir;
	int safe_symlinks;
	int copy_unsafe_links;
	int block_size;
	int size_only;
	int bwlimit;
	size_t bwlimit_writemax;
	int delete_after;
	int only_existing;
	int opt_ignore_existing;
	int max_delete;
	int ignore_errors;
	int modify_window;
	int blocking_io;
	int csum_length;
	int checksum_seed;
	int batch_fd;
	int batch_gen_fd;
	int write_batch;
	int read_batch;
	int inplace;
	int keep_dirlinks;
	int sanitize_paths;
	int list_only;
	int ssl;
	struct stats stats;
	struct exclude_list_struct exclude_list;
	struct exclude_list_struct local_exclude_list;
	struct exclude_list_struct server_exclude_list;
	char *exclude_path_prefix;	
	struct idlist *uidlist;
	struct idlist *gidlist;
	char *obuf;
	char *cbuf;
	char *dbuf;
	struct redo_list *redo_list_head;
	struct redo_list *redo_list_tail;
	struct msg_list *msg_list_head;
	struct msg_list *msg_list_tail;
	struct nwdelete_list *delete_list;
	int dlist_len;
	int dlist_alloc_len;
	char *cleanup_fname;
	char *cleanup_new_fname;
	struct file_struct *cleanup_file;
	int cleanup_fd_r;
	int cleanup_fd_w;
	pid_t cleanup_pid;
	struct map_struct *cleanup_buf;
	int sumresidue;
	char sumrbuf[NW_CSUM_CHUNK];
	char topsrcname[MAXPATHLEN];
	struct mdfour checksum_md;
	OFF_T last_match;
	int false_alarms;
	int tag_hits;
	int matches;
	int64 data_transfer;
	int total_false_alarms;
	int total_tag_hits;
	int total_matches;
	z_stream rx_strm;
	int rx_token;
	int rx_run;
	z_stream tx_strm;
	int last_token;
	int run_start;
	int last_run_end;
	enum { r_init, r_idle, r_running, r_inflating, r_inflated } recv_state;
	size_t remaining;
	char *backup_suffix;
	int backup_suffix_len;	
	char *backup_dir;
	int backup_dir_len;	
	char *tmpdir;
	char *compare_dest;
	char *flist_dir;
	int deletion_count;
	struct target *targets;
	size_t *tag_table;
	OFF_T  last_ofs;
	struct timeval print_time;
	struct timeval start_time;
	OFF_T  start_ofs;
	char TimeBuf[200];
	char last_byte;
	int last_sparse;
	WSAEVENT readEvent;
	WSAEVENT writeEvent;
	WSAEVENT errorEvent;
	service **ServicePtrs;
	int iNumServices;
	int iServiceIndex;
	BOOL bInGlobalSection;
	int suffix_specified;
	char *files_from;
	int link_dest;
	int filesfrom_fd;
	int implied_dirs;
//	int daemon_over_rsh;
	char *remote_filesfrom_file;
	int archive_mode;
	int eol_nulls;
	int io_filesfrom_f_in;
	int io_filesfrom_f_out;
	char io_filesfrom_buf[2048];
	char *io_filesfrom_bp;
	char io_filesfrom_lastchar;
	int io_filesfrom_buflen;
	char *batch_prefix;
	char *batch_name;
	int kludge_around_eof;
	const char *io_write_phase;
	const char *io_read_phase;
	int fdb;
	int fdb_delta;
	int fdb_open;
	int fdb_close;
	struct file_list *batch_flist;
	char empty_sum[MD4_SUM_LENGTH];
	unsigned int file_struct_len;
	int flist_dir_len;
	int daemon_opt;   
	int modify_window_set;
	struct progress_history ph_start;
	struct progress_history ph_list[PROGRESS_HISTORY_SECS];
	int newest_hpos;
	int oldest_hpos;
	size_t bufferIdx;
	size_t bufferSz;
	int f_csums;
	int f_delta;
	struct timeval prior_tv;
	long io_total_written; 
	char *wf_writeBuf;
	size_t wf_writeBufSize;
	size_t wf_writeBufCnt;
	struct file_list *received_flist;
	
	// renamed because they are statics
	char f_name_names[5][MAXPATHLEN];
	int f_name_n;
	char rf_file_sum1[MD4_SUM_LENGTH];
	char rf_file_sum2[MD4_SUM_LENGTH];
	char *checksum2_buf1;
	int checksum2_len1;
	int rdt_init_done;
	int rdt_saved_flag;
	int residue;
	char *srt_buf;
	int sdt_init_done;
	int sdt_flush_pending;
	char *mf_lastdir;
	int mf_lastdir_len;
	time_t sfe_modtime;
	mode_t sfe_mode;
	uint64 sfe_dev;
	dev_t sfe_rdev;
	uint32 sfe_rdev_major;
	uid_t sfe_uid;
	gid_t sfe_gid;
	char sfe_lastname[MAXPATHLEN];
	time_t rfe_modtime;
	mode_t rfe_mode;
	uint64 rfe_dev;
	dev_t rfe_rdev;
	uint32 rfe_rdev_major;
	uid_t rfe_uid;
	gid_t rfe_gid;
	char rfe_lastname[MAXPATHLEN];
	char *rfe_lastdir;
	int rfe_lastdir_len;
	char *sfl_lastdir;
	int sfl_lastdir_len;	
	char *get_s;
	int get_more;
	char as_user[100];
	int kb_initialised;
	char ca_addr_buf[100];
	int ca_initialised;
	char cn_name_buf[100];
	char cn_port_buf[100];
	int cn_initialised;
	int pd_initialised;
	struct mdfour *md4_m;
	uid_t mlu_last_in;
	uid_t mlu_last_out;
	gid_t mlg_last_in;
	gid_t mlg_last_out;

	int error_pipe[2];
	struct file_list *gFlist;
	time_t starttime;
	BOOL use_lfs;
	void *argContext;
	void *DmemRoot;
	void *trustBarSession;
	char uid_name[257];
	int  threadType;
	int generator_send_flag;
	semaphore* send_recv_semaphore;
	int logEntered;
	int inside_cleanup;
	int cleanup_got_literal;
	int recursionDepth;
	int depth;
	int lastdir_depth;
	int in_block;
	NXCpuId_t bound_cpu;
	struct sum_struct null_sum;	
	char *full_name_result;
	char *auth_user;
	int msg_fd_in;
	int msg_fd_out;

#ifdef SCOREBOARD
	ap_sb_handle_t * sb_handle;
#endif
} TSD;

/* a global variable */
extern pthread_key_t tsdKey;
#define get_tsd            TSD* tsd = (TSD*) pthread_getspecific(tsdKey);

#define io_multiplexing_out    	tsd->io_multiplexing_out
#define io_multiplexing_in     	tsd->io_multiplexing_in
#define multiplex_in_fd      	tsd->multiplex_in_fd
#define multiplex_out_fd        tsd->multiplex_out_fd
#define sock_f_in				tsd->sock_f_in
#define sock_f_out				tsd->sock_f_out
#define write_batch_monitor_in	tsd->write_batch_monitor_in
#define write_batch_monitor_out	tsd->write_batch_monitor_out
#define last_io				    tsd->last_io
#define no_flush				tsd->no_flush
#define iobuf_out				tsd->iobuf_out
#define iobuf_out_cnt			tsd->iobuf_out_cnt
#define iobuf_in				tsd->iobuf_in
#define iobuf_in_siz			tsd->iobuf_in_siz
#define iobuf_in_ndx			tsd->iobuf_in_ndx
#define curr_dir				tsd->curr_dir
#define curr_dir_len			tsd->curr_dir_len
#define full_name				tsd->full_name
#define backup_dir_buf			tsd->backup_dir_buf
#define backup_dir_remainder	tsd->backup_dir_remainder
#define log_got_error			tsd->log_got_error
#define verbose					tsd->verbose
#define quiet					tsd->quiet
#define make_backups			tsd->make_backups
#define whole_file				tsd->whole_file
#define no_whole_file			tsd->no_whole_file
#define copy_links				tsd->copy_links
#define preserve_links			tsd->preserve_links
#define preserve_hard_links		tsd->preserve_hard_links
#define preserve_perms			tsd->preserve_perms
#define preserve_devices		tsd->preserve_devices
#define preserve_uid			tsd->preserve_uid
#define preserve_gid			tsd->preserve_gid
#define preserve_times			tsd->preserve_times
#define update_only				tsd->update_only
#define cvs_exclude				tsd->cvs_exclude
#define dry_run					tsd->dry_run
#define local_server			tsd->local_server
#define ignore_times			tsd->ignore_times
#define delete_mode				tsd->delete_mode
#define delete_excluded			tsd->delete_excluded
#define one_file_system			tsd->one_file_system
#define protocol_version		tsd->protocol_version
#define remote_protocol			tsd->remote_protocol
#define sparse_files			tsd->sparse_files
#define do_compression			tsd->do_compression
#define compression_level		tsd->compression_level
#define am_root					tsd->am_root
#define orig_umask				tsd->orig_umask
#define relative_paths			tsd->relative_paths
#define numeric_ids				tsd->numeric_ids
#define force_delete			tsd->force_delete
#define io_timeout				tsd->io_timeout
#define io_error				tsd->io_error
#define read_only				tsd->read_only
#define module_id				tsd->module_id
#define am_server				tsd->am_server
#define am_sender				tsd->am_sender
#define am_generator			tsd->am_generator
#define daemon_over_rsh			tsd->daemon_over_rsh
#define recurse					tsd->recurse
#define do_stats				tsd->do_stats
#define do_progress				tsd->do_progress
#define keep_partial			tsd->keep_partial
#define select_timeout			tsd->select_timeout
#define partial_dir				tsd->partial_dir
#define safe_symlinks			tsd->safe_symlinks
#define copy_unsafe_links		tsd->copy_unsafe_links
#define block_size				tsd->block_size
#define size_only				tsd->size_only
#define bwlimit					tsd->bwlimit
#define bwlimit_writemax		tsd->bwlimit_writemax
#define delete_after			tsd->delete_after
#define only_existing			tsd->only_existing
#define opt_ignore_existing		tsd->opt_ignore_existing
#define max_delete				tsd->max_delete
#define ignore_errors			tsd->ignore_errors
#define modify_window			tsd->modify_window
#define blocking_io				tsd->blocking_io
#define csum_length				tsd->csum_length
#define checksum_seed			tsd->checksum_seed
#define batch_fd				tsd->batch_fd
#define batch_gen_fd			tsd->batch_gen_fd
#define write_batch				tsd->write_batch
#define read_batch				tsd->read_batch
#define inplace					tsd->inplace
#define in_block				tsd->in_block
#define lastdir_depth			tsd->lastdir_depth
#define keep_dirlinks			tsd->keep_dirlinks
#define always_checksum			tsd->always_checksum
#define sanitize_paths			tsd->sanitize_paths
#define list_only				tsd->list_only
#define ssl						tsd->ssl
#define stats					tsd->stats
#define uidlist					tsd->uidlist
#define gidlist					tsd->gidlist
#define obuf					tsd->obuf
#define cbuf					tsd->cbuf
#define dbuf					tsd->dbuf
#define redo_list_head			tsd->redo_list_head
#define redo_list_tail			tsd->redo_list_tail
#define msg_list_head			tsd->msg_list_head
#define msg_list_tail			tsd->msg_list_tail
#define delete_list				tsd->delete_list
#define dlist_len				tsd->dlist_len
#define dlist_alloc_len			tsd->dlist_alloc_len
#define cleanup_fname			tsd->cleanup_fname
#define cleanup_new_fname		tsd->cleanup_new_fname
#define cleanup_file			tsd->cleanup_file
#define cleanup_fd_r			tsd->cleanup_fd_r
#define cleanup_fd_w			tsd->cleanup_fd_w
#define cleanup_pid				tsd->cleanup_pid
#define cleanup_buf				tsd->cleanup_buf
#define sumresidue				tsd->sumresidue
#define sumrbuf					tsd->sumrbuf
#define last_match				tsd->last_match
#define false_alarms			tsd->false_alarms
#define tag_hits				tsd->tag_hits
#define data_transfer			tsd->data_transfer
#define total_false_alarms		tsd->total_false_alarms
#define total_tag_hits			tsd->total_tag_hits
#define total_matches			tsd->total_matches
#define rx_token				tsd->rx_token
#define rx_run					tsd->rx_run
#define rx_strm					tsd->rx_strm
#define tx_strm					tsd->tx_strm
#define last_token				tsd->last_token
#define run_start				tsd->run_start
#define last_run_end			tsd->last_run_end
#define recv_state				tsd->recv_state
#define remaining				tsd->remaining
#define local_exclude_list		tsd->local_exclude_list
#define exclude_list			tsd->exclude_list
#define server_exclude_list		tsd->server_exclude_list
#define exclude_path_prefix		tsd->exclude_path_prefix
#define topsrcname				tsd->topsrcname
#define residue					tsd->residue
#define backup_suffix			tsd->backup_suffix
#define backup_suffix_len		tsd->backup_suffix_len
#define backup_dir				tsd->backup_dir
#define backup_dir_len			tsd->backup_dir_len
#define tmpdir					tsd->tmpdir
#define compare_dest			tsd->compare_dest
#define flist_dir				tsd->flist_dir
#define deletion_count			tsd->deletion_count
#define exclude_list			tsd->exclude_list
#define targets					tsd->targets
#define tag_table				tsd->tag_table
#define last_ofs				tsd->last_ofs
#define print_time				tsd->print_time
#define start_time				tsd->start_time
#define start_ofs				tsd->start_ofs
#define TimeBuf					tsd->TimeBuf
#define last_byte				tsd->last_byte
#define last_sparse				tsd->last_sparse
#define bInGlobalSection		tsd->bInGlobalSection
#define iNumServices			tsd->iNumServices
#define ServicePtrs				tsd->ServicePtrs
#define iServiceIndex			tsd->iServiceIndex
#define starttime				tsd->starttime
#define use_lfs					tsd->use_lfs
#define DmemRoot				tsd->DmemRoot
#define generator_send_flag		tsd->generator_send_flag
#define send_recv_semaphore		tsd->send_recv_semaphore
#define suffix_specified		tsd->suffix_specified
#define files_from				tsd->files_from
#define link_dest				tsd->link_dest
#define filesfrom_fd			tsd->filesfrom_fd
#define implied_dirs			tsd->implied_dirs
#define f_csums					tsd->f_csums
#define f_delta					tsd->f_delta
#define prior_tv				tsd->prior_tv
#define remote_filesfrom_file	tsd->remote_filesfrom_file
#define archive_mode			tsd->archive_mode
#define eol_nulls				tsd->eol_nulls
#define io_filesfrom_f_in		tsd->io_filesfrom_f_in
#define io_filesfrom_f_out		tsd->io_filesfrom_f_out
#define io_filesfrom_buf		tsd->io_filesfrom_buf
#define io_filesfrom_bp			tsd->io_filesfrom_bp
#define io_filesfrom_lastchar	tsd->io_filesfrom_lastchar
#define io_filesfrom_buflen		tsd->io_filesfrom_buflen
#define batch_prefix			tsd->batch_prefix
#define batch_name				tsd->batch_name
#define kludge_around_eof		tsd->kludge_around_eof
#define io_write_phase			tsd->io_write_phase
#define io_read_phase			tsd->io_read_phase
#define inside_cleanup			tsd->inside_cleanup
#define cleanup_got_literal		tsd->cleanup_got_literal
#define null_sum				tsd->null_sum
#define batch_flist				tsd->batch_flist
#define auth_user				tsd->auth_user
#define msg_fd_in				tsd->msg_fd_in
#define msg_fd_out				tsd->msg_fd_out
#define empty_sum				tsd->empty_sum
#define file_struct_len			tsd->file_struct_len
#define flist_dir_len			tsd->flist_dir_len
#define daemon_opt				tsd->daemon_opt
#define modify_window_set		tsd->modify_window_set
#define ph_start				tsd->ph_start
#define ph_list					tsd->ph_list
#define newest_hpos				tsd->newest_hpos
#define oldest_hpos				tsd->oldest_hpos
#define bufferIdx				tsd->bufferIdx
#define bufferSz				tsd->bufferSz
#define wf_writeBuf				tsd->wf_writeBuf
#define wf_writeBufSize			tsd->wf_writeBufSize
#define wf_writeBufCnt			tsd->wf_writeBufCnt
#define received_flist			tsd->received_flist

#define    	REGISTRYKEY_RSYNC				"Rsync\\Volumes"
#define    	REGISTRYKEY_RSYNC_SYS			"Rsync\\Daemon"
#define		REGISTRYVALUE_ENABLED			"Enabled"
#define		REGISTRYVALUE_AGENT_NDS			"NDS Agent Discovery"
#define		REGISTRYVALUE_AGENT_SLP			"SLP Agent Discovery"
#define 	REGISTRYVALUE_ADDRESS			"SMDR Address";
#define		REGISTRYVALUE_STARTTIME			"StartTime"
#define		REGISTRYVALUE_ENDTIME			"EndTime"
#define		REGISTRYVALUE_TIME				"Time"
#define		REGISTRYVALUE_STATUS			"Status"
#define		REGISTRYVALUE_TOTALWRITE 		"TotalWrite"
#define		REGISTRYVALUE_TOTALREAD 		"TotalRead"
#define		REGISTRYVALUE_TOTALSIZE 		"TotalSize"
#define		REGISTRYVALUE_TOTAL_XFER_SIZE	"TotalTransferSize"
#define		REGISTRYVALUE_NUM_XFER_FILES	"NumTransferFiles"
#define		REGISTRYVALUE_ERRSTRING			"ErrorString"
#define		REGISTRYVALUE_FILE				"File"
#define		REGISTRYVALUE_LINE				"Line"
#define		REGISTRYVALUE_FILELISTCOUNT		"FileListCount" 		// count of files in the file list
#define		REGISTRYVALUE_FILELISTSIZE		"FileListSize"			// number of bytes transferred so far
#define		REGISTRYVALUE_FILELISTTOTALSIZE	"FileListTotalSize"		// total size of the files in the file list
#define		REGISTRYVALUE_FILELISTINDEX		"FileListIndex"			// file number we are transferring right now
#define		REGISTRYVALUE_THROUGHPUT		"Throughput"			// throughput for the last run
#define		REGISTRYVALUE_RSYNC_ERROR 		"RsyncError"

#define		MAX_RECURSION_DEPTH				35

typedef struct threadParams
{
	int (*fn)(int,int );
	int f_in;
	int f_out;
	TSD tsd;
} netwareThreadParams;

typedef struct threadRecvParams
{
	int(*fn)(int ,struct file_list *,char * );
	void (*report)(int);
	int f_in;
	int f_out;
	int error_pipe[2];
	char local_name[MAXPATHLEN];
	struct file_list *flist;
	TSD tsd;
} netwareThreadRecvParams;

#if 0
#define malloc(size) _nwmalloc((size), __FILE__, __LINE__)
#define free(pointer) _nwfree((pointer), __FILE__, __LINE__)
#define strdup(ptr) _nwstrdup((ptr), __FILE__, __LINE__)
#define realloc(ptr, size) _nwrealloc( (ptr), (size), __FILE__, __LINE__)
#define calloc(n, size) _nwcalloc( (n), (size), __FILE__, __LINE__)

void *_nwmalloc(size_t size, const char* file, int line);
void *_nwcalloc(size_t nmemb, size_t size, const char* file, int line);
void *_nwrealloc(void *ID, size_t size, const char* file, int line);
void _nwfree(void *ID, const char* file, int line);
char *_nwstrdup(const char * str, const char* file, int line);
#endif


//#define RSYNC_SCREEN_NAME "Rsync Client 2.6.0"
#define RSYNC_DAEMON_SCREEN_NAME "Rsync 2.6.3"

void set_linger(int fd, int onOff);

#define nwexit_cleanup(code) _nwexit_cleanup(code, __FILE__, __LINE__)

//#define WS2_32_closesocket(socket) _nwWS2_32_closesocket(socket, __FILE__, __LINE__)
//#define WS2_32_socket(domain, type, protocol) _nwWS2_32_socket(domain, type, protocol, __FILE__, __LINE__)

thread *create_thread(void (thread_fn)(void *), void *thread_arg);
thread *create_threadgroup(void (thread_fn)(void *), void *thread_arg);
int kill_thread(thread *thread_id);
void threadCleanup(int flag);
void free_services();

semaphore *create_semaphore(int initial);
int acquire_semaphore(semaphore *semaphore_id);
int release_semaphore(semaphore *semaphore_id);
int destroy_semaphore(semaphore *semaphore_id);

semaphore *create_namedsemaphore(const char* semaphoreName, int initialValue);
void destroy_namedsemaphore(semaphore *semaphore_id);
int examine_namedsemaphore(semaphore *semaphore_id);

void netwareForkThread(void* arg);
void netwareRecvThread(void* arg);

int GetStatusFromRegistry(int *pStatus);
int SaveStatusToRegistry(int code, BOOL bCleanAll);
int SaveTimeToRegistry(int code, char *regKey);
int SaveStatsToRegistry(int code);
int	SaveFileLineToRegistry(const char* file, int line);
int SaveFileListCountToRegistry(struct file_list *flist);
int SaveFileListIndexToRegistry(const char *op, struct file_list *flist, int index, int fileCount);

extern scr_t myErrorScreen;

#else
extern int verbose;
#endif

#ifndef HAVE_INET_NTOP
const char *                 
inet_ntop(int af, const void *src, char *dst, size_t size);
#endif /* !HAVE_INET_NTOP */

#ifndef HAVE_INET_PTON
int inet_pton(int af, const char *src, void *dst);
#endif

#ifdef MAINTAINER_MODE
const char *get_panic_action(void);
#endif

#define UNUSED(x) x __attribute__((__unused__))

#ifndef __NETWARE__
extern const char *io_write_phase, *io_read_phase;
#endif